open Base
open Printf
open Syslog
open Lwt.Infix
open Okweb

module Daemon = Okauth_server.Daemon.Make(struct
    let db_path = Okauth_server.default_db_path
    let log msg = syslog msg
  end)

let log_request_response req resp =
  let remote_host = match req |> Request.get_client_addr with
    | Some (Unix.ADDR_INET (inet_addr, _)) ->
      Core.Unix.Inet_addr.to_string inet_addr
    | Some (Unix.ADDR_UNIX pipe) ->
      pipe
    | None ->
      "<unknown host>"
  in
  let zone = Lazy.force Core.Time.Zone.local in
  let req_end_time = Core.Time.(now () |> to_string_iso8601_basic ~zone) in
  let http_method = Request.get_http_method req in
  let path = req |> Request.get_uri |> Uri.path_and_query in
  let resp_code = Response.get_status resp in
  let content_length = resp |> Response.get_body |> Body.length |> Option.value ~default:(-1) in
  sprintf "%s - - [%s] \"%s %s\" %d %d" remote_host req_end_time http_method path resp_code content_length |> syslog

let handle_error ex req =
  let exn_uuid = Uuid_unix.create () |> Uuid.to_string in
  sprintf "%s: %s" exn_uuid (Exn.to_string ex) |> syslog;
  let body = sprintf "An exception occurred: %s. Please check logs." exn_uuid in
  let resp = Response.(
      of_status 500
      |> with_header ~key:"X-Exception-ID" ~value:exn_uuid
      |> with_string_body body) in
  log_request_response req resp;
  Lwt.return resp

let handler req =
  try
    Daemon.process_request req >|= fun resp ->
    log_request_response req resp;
    resp
  with
  | ex -> handle_error ex req

let get_host_port () =
  let open Core in
  let from_env (type t) (module T: Stringable with type t = t) ~key ~default =
    let value_opt = Sys.getenv key in
    try
      value_opt |> Option.map ~f:T.of_string |> Option.value ~default
    with
    | Failure _ ->
      sprintf "Invalid %s: %s. Using %s instead" key (Option.value value_opt ~default:"") (T.to_string default) |> syslog;
      default
  in
  let host = from_env (module String) ~key:"HOST" ~default:"localhost" in
  let port = from_env (module Int) ~key:"PORT" ~default:4243 in
  (host, port)

let () =
  openlog ~id:"okauthd" ~options:[ Open_option.PID; Open_option.PERROR; ] ();
  match Daemon.init () with
  | Result.Error err ->
    syslog err;
    Caml.exit 1
  | Result.Ok _ ->
    let (address, port) = get_host_port () in
    sprintf "Starting OkAuth Daemon v%s on %s:%d" [%blob "../../version"] address port |> syslog;
    Lwt_main.run (Server.listen ~address ~port handler);
